use core::{
    cmp::Ordering,
    fmt::Debug,
    ops::{Bound, RangeBounds},
};

use crate::Error;

pub mod error {
    use thiserror::Error;

    #[derive(Debug, Error)]
    pub enum Error {
        #[error("read failed")]
        ReadFailed,
        #[error("invalid magic signature")]
        InvalidMagic,
        #[error("invalid offset")]
        InvalidOffset,
        #[error("Expected an internal node")]
        ExpectedInternalNode,
        #[error("Expected a leaf node")]
        ExpectedLeafNode,
        #[error("Failed to read with logical offset")]
        BadLogicalAddress,
        #[error("Default subvolume root could not be found")]
        NoDefaultSubvolRoot,
        #[error("Default subvolume root fs tree could not be found")]
        NoDefaultSubvolFsRoot,
        #[error("INode could not be found in FsTree")]
        INodeNotFound,
        #[error("attempted to access {index}th item out of bounds {range:?}")]
        OutOfBounds {
            range: core::ops::Range<usize>,
            index: usize,
        },
        #[error("Invalid checksum: expected {expected:#?} but got {actual:#?}")]
        InvalidChecksum {
            expected: [u8; 32],
            actual: [u8; 32],
        },
        #[error("{0}")]
        ScrollError(scroll::Error),
    }

    impl From<scroll::Error> for Error {
        fn from(value: scroll::Error) -> Self {
            Self::ScrollError(value)
        }
    }

    pub type Result<T> = core::result::Result<T, Error>;
}

pub trait Read {
    fn read(&self, dst: &mut [u8], address: u64) -> error::Result<()>;
    fn alignment(&self) -> usize {
        1
    }
}

#[cfg(all(any(feature = "std", test), unix))]
impl Read for std::fs::File {
    fn read(&self, dst: &mut [u8], address: u64) -> error::Result<()> {
        use std::os::unix::prelude::FileExt;
        self.read_at(dst, address).map_err(|_| Error::ReadFailed)?;
        Ok(())
    }
}

pub mod file;
pub mod tree;
pub mod volume;

pub fn cmp_start_bound<T: PartialOrd>(
    rhs: &core::ops::Bound<T>,
    lhs: &core::ops::Bound<T>,
) -> Option<core::cmp::Ordering> {
    match rhs {
        core::ops::Bound::Included(r) => match lhs {
            core::ops::Bound::Included(l) => r.partial_cmp(&l),
            core::ops::Bound::Excluded(l) => match r.partial_cmp(&l) {
                Some(core::cmp::Ordering::Equal) => Some(core::cmp::Ordering::Less),
                i => i,
            },
            core::ops::Bound::Unbounded => Some(core::cmp::Ordering::Greater),
        },
        core::ops::Bound::Excluded(r) => match lhs {
            core::ops::Bound::Excluded(l) => r.partial_cmp(&l),
            core::ops::Bound::Included(l) => match r.partial_cmp(&l) {
                Some(core::cmp::Ordering::Equal) => Some(core::cmp::Ordering::Greater),
                i => i,
            },
            core::ops::Bound::Unbounded => Some(core::cmp::Ordering::Greater),
        },
        core::ops::Bound::Unbounded => match lhs {
            core::ops::Bound::Unbounded => Some(core::cmp::Ordering::Equal),
            _ => Some(core::cmp::Ordering::Less),
        },
    }
}

pub fn cmp_end_bound<T: PartialOrd>(
    rhs: &core::ops::Bound<T>,
    lhs: &core::ops::Bound<T>,
) -> Option<core::cmp::Ordering> {
    match rhs {
        core::ops::Bound::Included(r) => match lhs {
            core::ops::Bound::Included(l) => r.partial_cmp(&l),
            core::ops::Bound::Excluded(l) => match r.partial_cmp(&l) {
                Some(core::cmp::Ordering::Equal) => Some(core::cmp::Ordering::Greater),
                i => i,
            },
            core::ops::Bound::Unbounded => Some(core::cmp::Ordering::Less),
        },
        core::ops::Bound::Excluded(r) => match lhs {
            core::ops::Bound::Excluded(l) => r.partial_cmp(&l),
            core::ops::Bound::Included(l) => match r.partial_cmp(&l) {
                Some(core::cmp::Ordering::Equal) => Some(core::cmp::Ordering::Greater),
                i => i,
            },
            core::ops::Bound::Unbounded => Some(core::cmp::Ordering::Less),
        },
        core::ops::Bound::Unbounded => match lhs {
            core::ops::Bound::Unbounded => Some(core::cmp::Ordering::Equal),
            _ => Some(core::cmp::Ordering::Greater),
        },
    }
}

#[derive(Debug, PartialEq, Eq)]
struct StartBound<T: PartialOrd>(Bound<T>);
impl<T: PartialOrd> PartialOrd for StartBound<T> {
    fn partial_cmp(&self, other: &Self) -> Option<core::cmp::Ordering> {
        cmp_start_bound(&self.0, &other.0)
    }
}
impl<T: Ord> Ord for StartBound<T> {
    fn cmp(&self, other: &Self) -> core::cmp::Ordering {
        self.partial_cmp(other).unwrap()
    }
}

#[derive(Debug, PartialEq, Eq)]
struct EndBound<T: PartialOrd>(Bound<T>);

impl<T: PartialOrd> PartialOrd for EndBound<T> {
    fn partial_cmp(&self, other: &Self) -> Option<core::cmp::Ordering> {
        cmp_end_bound(&self.0, &other.0)
    }
}
impl<T: Ord> Ord for EndBound<T> {
    fn cmp(&self, other: &Self) -> core::cmp::Ordering {
        // safety: partial_cmp only returns None when returning T::partial_cmp
        self.partial_cmp(other).unwrap()
    }
}

pub fn range_countains_bound<T: Ord + Debug, R: RangeBounds<T>>(
    range: R,
    bound: Bound<&T>,
) -> bool {
    let start = &StartBound(bound);
    let end = &EndBound(bound);
    let r_start = &StartBound(range.start_bound());
    let r_end = &EndBound(range.end_bound());
    log::info!(
        "start: {start:?} <=> {r_start:?}: {:?} {:?}",
        start.cmp(r_start),
        r_start.cmp(start)
    );
    log::info!(
        "end: {end:?} <=> {r_end:?}: {:?} {:?}",
        end.cmp(r_end),
        r_end.cmp(end)
    );
    (start.cmp(r_start).is_ge() || r_start.cmp(start).is_le())
        && (end.cmp(r_end).is_ge() || r_end.cmp(end).is_le())
}

#[cfg(test)]
mod bound_ord_tests {
    use core::{cmp::Ordering, ops::RangeBounds};

    use super::*;
    use test_log::test;

    #[test]
    fn start_bound_ord() {
        assert_eq!(
            cmp_start_bound(&Bound::Unbounded, &Bound::Included(0)),
            Some(Ordering::Less)
        );
        assert_eq!(
            cmp_start_bound::<i32>(&Bound::Unbounded, &Bound::Unbounded),
            Some(Ordering::Equal)
        );
        assert_eq!(
            cmp_start_bound(&Bound::Included(0), &Bound::Included(0)),
            Some(Ordering::Equal)
        );
        assert_eq!(
            cmp_start_bound(&Bound::Excluded(0), &Bound::Included(0)),
            Some(Ordering::Greater)
        );
        assert_ne!(
            cmp_start_bound(&Bound::Excluded(0), &Bound::Included(1)),
            Some(Ordering::Greater)
        );
        // This is actually WRONG and is why we have to test both ways because I
        // can't think of a way to actually determine how the 2 bounds are
        // ordered without knowing the smallest discrete step size of T.
        //
        // In this case technically they should be equal, but for floats (which
        // arent ord anyways?) this would be wrong
        assert_eq!(
            cmp_start_bound(&Bound::Included(1), &Bound::Excluded(0)),
            Some(Ordering::Greater)
        );

        assert_eq!(
            cmp_start_bound(&Bound::Included(0), &Bound::Excluded(1)),
            Some(Ordering::Less)
        );
    }

    #[test]
    fn end_bound_ord() {
        assert_eq!(
            cmp_end_bound::<i32>(&Bound::Unbounded, &Bound::Unbounded),
            Some(Ordering::Equal)
        );
        assert_eq!(
            cmp_end_bound(&Bound::Unbounded, &Bound::Included(0)),
            Some(Ordering::Greater)
        );
        assert_eq!(
            cmp_end_bound(&Bound::Included(0), &Bound::Included(0)),
            Some(Ordering::Equal)
        );
        assert_eq!(
            cmp_end_bound(&Bound::Excluded(0), &Bound::Included(0)),
            Some(Ordering::Greater)
        );
        assert_ne!(
            cmp_end_bound(&Bound::Excluded(0), &Bound::Included(1)),
            Some(Ordering::Greater)
        );
        // This is actually WRONG and is why we have to test both ways because I
        // can't think of a way to actually determine how the 2 bounds are
        // ordered without knowing the smallest discrete step size of T.
        //
        // In this case technically they should be equal, but for floats (which
        // arent ord anyways?) this would be wrong
        assert_eq!(
            cmp_end_bound(&Bound::Included(1), &Bound::Excluded(0)),
            Some(Ordering::Greater)
        );

        assert_eq!(
            cmp_end_bound(&Bound::Included(0), &Bound::Excluded(1)),
            Some(Ordering::Less)
        );
    }

    #[test]
    fn test_bound_ord() {
        let r1 = 0..4;
        let r2 = 2..3;

        assert_eq!(
            cmp_start_bound(&r1.start_bound(), &r2.start_bound()),
            Some(core::cmp::Ordering::Less)
        );
        assert_eq!(
            cmp_end_bound(&r1.end_bound(), &r2.end_bound()),
            Some(core::cmp::Ordering::Greater)
        );

        assert_eq!(
            cmp_start_bound(&r2.start_bound(), &r1.start_bound()),
            Some(core::cmp::Ordering::Greater)
        );
        assert_eq!(
            cmp_end_bound(&r2.end_bound(), &r1.end_bound()),
            Some(core::cmp::Ordering::Less)
        );

        let r1 = 0..=8;
        let r2 = 0..9;

        assert_eq!(
            cmp_start_bound(&r1.start_bound(), &r2.start_bound()),
            Some(core::cmp::Ordering::Equal)
        );
        assert_eq!(
            cmp_end_bound(&r1.end_bound(), &r2.end_bound()),
            Some(core::cmp::Ordering::Less)
        );
        assert_eq!(
            cmp_end_bound(&r2.end_bound(), &r1.end_bound()),
            Some(core::cmp::Ordering::Greater)
        );
    }
}

pub fn ranges_intersect<T, R1, R2>(rhs: R1, lhs: R2) -> ()
where
    T: Ord,
    R1: core::ops::RangeBounds<T>,
    R2: core::ops::RangeBounds<T>,
{
    let a = rhs.start_bound();
    let b = rhs.end_bound();
    let x = lhs.start_bound();
    let y = lhs.end_bound();

    // check that a <=> x is different than b <=> x

    // a <=> x
    match a {
        Bound::Included(a) => match x {
            Bound::Included(x) => a.partial_cmp(x),
            Bound::Excluded(x) => match a.partial_cmp(x) {
                Some(Ordering::Equal) => Some(Ordering::Less),
                ord => ord,
            },
            Bound::Unbounded => Some(Ordering::Greater),
        },
        Bound::Excluded(a) => match x {
            Bound::Included(x) => match a.partial_cmp(x) {
                Some(Ordering::Equal) => Some(Ordering::Greater),
                ord => ord,
            },
            Bound::Excluded(x) => a.partial_cmp(x),
            Bound::Unbounded => Some(Ordering::Less),
        },
        Bound::Unbounded => match x {
            Bound::Unbounded => Some(Ordering::Equal),
            _ => Some(Ordering::Less),
        },
    };

    todo!()
}
